/*
	$Id: FileDescriptor.c 104 2006-03-02 15:35:08Z syl $

	Copyright (C) 2006 Syllabub
	Maroontress Fast Software.
*/

#include <stdlib.h>
#include "tricks.h"
#include "FileDescriptor.h"
#include "impl/InputStream.h"
#include "impl/OutputStream.h"

/* __final__ */ struct impl_FDInputStream {
    union {
	struct impl_InputStream inputStream;
    } super;
    int fd;
    int autoClosing;
};

union FDIS {
    struct impl_InputStream *inputStream;
    struct impl_FDInputStream *impl;
} TRANSPARENT_UNION;

static ssize_t
FDIS_read(union FDIS this, void *data, size_t size)
{
    struct impl_FDInputStream *impl = this.impl;
    int fd = impl->fd;

    return read(fd, data, size);
}

static int
FDIS_close(union FDIS this)
{
    struct impl_FDInputStream *impl = this.impl;
    int s;

    s = (impl->autoClosing) ? close(impl->fd) : 0;
    free(impl);
    return s;
}

union InputStream
FileDescriptor_createInputStream(int fd, int autoClosing)
{
    struct impl_FDInputStream *impl;

    if ((impl = malloc(sizeof(*impl))) == NULL) {
	return (union InputStream)null;
    }
    impl->super.inputStream.read = FDIS_read;
    impl->super.inputStream.close = FDIS_close;
    impl->fd = fd;
    impl->autoClosing = autoClosing;
    return (union InputStream)&impl->super.inputStream;
}

/* __final__ */ struct impl_FDOutputStream {
    union {
	struct impl_OutputStream outputStream;
    } super;
    int fd;
    int autoClosing;
};

union FDOS {
    struct impl_OutputStream *outputStream;
    struct impl_FDOutputStream *impl;
} TRANSPARENT_UNION;

static ssize_t
FDOS_write(union FDOS this, const void *data, size_t size)
{
    struct impl_FDOutputStream *impl = this.impl;
    int fd = impl->fd;

    return write(fd, data, size);
}

static int
FDOS_close(union FDOS this)
{
    struct impl_FDOutputStream *impl = this.impl;
    int s;

    s = (impl->autoClosing) ? close(impl->fd) : 0;
    free(impl);
    return s;
}

union OutputStream
FileDescriptor_createOutputStream(int fd, int autoClosing)
{
    struct impl_FDOutputStream *impl;

    if ((impl = malloc(sizeof(*impl))) == NULL) {
        return (union OutputStream)null;
    }
    impl_OutputStream_new(&impl->super.outputStream,
			  FDOS_write,
			  FDOS_close);
    impl->fd = fd;
    impl->autoClosing = autoClosing;
    return (union OutputStream)&impl->super.outputStream;
}
