/*
	$Id: Test.c 191 2006-05-06 20:46:49Z syl $

	Copyright (C) 2006 Syllabub
	Maroontress Fast Software.
*/

#ifdef linux
#include <alloca.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <err.h>
#include <dlfcn.h>
#include <bfd.h>
#include "Auditor.h"
#include "TestListener.h"
#include "Test.h"

static void
findAllMethods(const char *file, bfd *abfd, struct TestListener *listener)
{
    asection *s;
    asymbol **sy = (asymbol **) NULL;
    bfd *sybfd;
    int k, storage, symcount;
    const char *name, *section;

    if ((storage = bfd_get_dynamic_symtab_upper_bound(abfd)) < 0)
	errx(1, "bfd_get_dynamic_symtab_upper_bound() failed: %s", file);
    else if (storage > 0)
	sy = (asymbol **)alloca(storage);

    if ((symcount = bfd_canonicalize_dynamic_symtab(abfd, sy)) < 0)
	errx(1, "bfd_canonicalize_dynamic_symtab() failed: %s", file);
    else if (symcount == 0)
	errx(1, "no dynamic symbols: %s", file);

    for (k = 0; k < symcount; ++sy, ++k) {
	if (*sy == NULL)
	    continue;
	if ((sybfd = bfd_asymbol_bfd(*sy)) == NULL)
	    continue;
	name = bfd_asymbol_name(*sy);
	s = bfd_get_section(*sy);
	section = bfd_get_section_name(sybfd, s);

	if (strcmp(section, ".audit") != 0
	    || ((*sy)->flags & BSF_FUNCTION) == 0) {
	    continue;
	}
	listener->methodFound(name);
    }
}

void
Test_findClass(const char *file, struct TestListener *listener)
{
    bfd *abfd;
    const char *desc;
    char **matching;
    void *h;

    bfd_init();

    if ((h = dlopen(file, RTLD_NOW)) == NULL) {
	warnx("%s", dlerror());
	warnx("%s: ignored.", file);
	return;
    }
    desc = (const char *)dlsym(h, "tf_description");
    listener->classOpened(file, desc);

    if ((abfd = bfd_openr(file, NULL)) == NULL)
	errx(1, "bfd_openr() failed: %s", file);
    if (bfd_check_format(abfd, bfd_archive))
	errx(1, "bfd_check_format() failed: %s", file);
    if (!bfd_check_format_matches(abfd, bfd_object, &matching))
	errx(1, "bfd_check_format_matches() failed: %s", file);
    findAllMethods(file, abfd, listener);
    bfd_close(abfd);

    listener->classClosed();
    (void)dlclose(h);
}

void
Test_invokeAll(const char *const *av)
{
    auto void classOpened(const char *, const char *);
    auto void classClosed(void);
    auto void methodFound(const char *);
    struct TestListener listener = {
        .classOpened = classOpened,
        .classClosed = classClosed,
        .methodFound = methodFound};
    struct {
	const char *className;
    } this;

    void classOpened(const char *className, const char *desc UNUSED) {
	this.className = className;
    }

    void methodFound(const char *methodName) {
	if (Auditor_invoke(this.className, methodName) != 0)
	    exit(1);
    }

    void classClosed(void) {
	this.className = NULL;
    }

    this.className = NULL;
    for (; *av != NULL; ++av) {
        Test_findClass(*av, &listener);
    }
}
